#include "RobotUnitReader.h"

#include "aron_conversions.h"

#include <RobotAPI/libraries/aron/core/navigator/data/container/Dict.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/Long.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/String.h>

#include <RobotAPI/libraries/RobotUnitDataStreamingReceiver/RobotUnitDataStreamingReceiver.h>

#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>

#include <SimoxUtility/algorithm/string/string_tools.h>

#include <istream>
#include <filesystem>
#include <fstream>


namespace armarx::armem::server::robot_state
{

    std::map<std::string, std::string> RobotUnitReader::readConfig(const std::string& configPath)
    {
        ARMARX_CHECK(std::filesystem::is_regular_file(configPath));

        std::map<std::string, std::string> config;

        std::ifstream in(configPath);
        std::string line;
        while (std::getline(in, line))
        {
            if (line.size() > 0)
            {
                if (simox::alg::starts_with(line, "//"))
                {
                    continue;
                }

                std::vector<std::string> mappings = simox::alg::split(line, ",");
                for (const auto& mapping : mappings)
                {
                    std::vector<std::string> split = simox::alg::split(mapping, "=>");
                    if (split.size() < 2)
                    {
                        ARMARX_WARNING_S << "A mapping in the RobotUnitConfig file is not valid. Ignoring it: " << mapping;
                        continue;
                    }
                    for (unsigned int i = 0; i < split.size() - 1; ++i)
                    {
                        ARMARX_DEBUG_S << "Setting sensor value '" << split[i] << "' to group '" << split[split.size() - 1] << "'.";
                        config.emplace(split[i], split[split.size() - 1]);
                    }
                }
            }
        }

        return config;
    }


    void RobotUnitReader::connect(
        armarx::plugins::RobotUnitComponentPlugin& robotUnitPlugin,
        armarx::plugins::DebugObserverComponentPlugin& debugObserverPlugin)
    {
        {
            config.loggingNames.push_back(properties.sensorPrefix);
            receiver = robotUnitPlugin.startDataSatreming(config);
            description = receiver->getDataDescription();
        }
        {
            // Thread-local copy of debug observer helper.
            debugObserver = DebugObserverHelper(
                                Logging::tag.tagName, debugObserverPlugin.getDebugObserver(), true
                            );
        }
    }


    void RobotUnitReader::run(
        float pollFrequency,
        std::queue<RobotUnitData>& dataQueue,
        std::mutex& dataMutex)
    {
        CycleUtil cycle(static_cast<int>(1000.f / pollFrequency));
        while (task and not task->isStopped())
        {
            if (std::optional<RobotUnitData> data = fetchAndGroupLatestRobotUnitData())
            {
                std::lock_guard g{dataMutex};
                dataQueue.push(data.value());
            }

            if (debugObserver)
            {
                debugObserver->sendDebugObserverBatch();
            }
            cycle.waitForCycleDuration();
        }
    }


    std::optional<RobotUnitData> RobotUnitReader::fetchAndGroupLatestRobotUnitData()
    {
        const std::optional<RobotUnitDataStreaming::TimeStep> data = fetchLatestData();
        if (not data.has_value())
        {
            return std::nullopt;
        }

        ARMARX_DEBUG << "RobotUnitData: Generating new aron map for current timestep";
        auto start = std::chrono::high_resolution_clock::now();

        RobotUnitData convertedAndGroupedData;
        for (const auto& [nameEntry, dataEntry] : description.entries)
        {
            std::string name = nameEntry;
            std::string jointOrWhateverName = nameEntry;

            std::string groupName = "";
            if (auto it = configSensorMapping.find(name); it != configSensorMapping.end())
            {
                groupName = it->second;
                jointOrWhateverName = it->second; // ???
            }
            else
            {
                const size_t first_dot_pos = name.find(".");
                const size_t second_dot_pos = name.find(".", first_dot_pos + 1); // find second occurence of "."
                if (second_dot_pos == std::string::npos)
                {
                    ARMARX_WARNING << "Could not find a groupname for the sensor with name " << name << ". "
                                   << "All sensors must be called sens.X.Y where X is the name of the group "
                                   << "and Y is the actual sensor. Ignoring this sensor.";
                    continue;
                }

                groupName = name.substr(0, second_dot_pos);
                jointOrWhateverName = name.substr(first_dot_pos + 1, second_dot_pos - first_dot_pos - 1);
                name = name.substr(second_dot_pos + 1); // remove the groupName, TODO check if +1 is valid
            }

            auto it = convertedAndGroupedData.groups.find(groupName);
            if (it == convertedAndGroupedData.groups.end())
            {
                namespace adn = aron::datanavigator;

                // Generate new dict for the group
                RobotUnitData::RobotUnitDataGroup group;
                group.timestamp = armem::Time::microSeconds(data->timestampUSec);
                group.name = groupName;
                group.data = std::make_shared<adn::DictNavigator>();

                group.data->addElement("EncoderGroupName", std::make_shared<adn::StringNavigator>(groupName));
                group.data->addElement("IterationId", std::make_shared<adn::LongNavigator>(data->iterationId));
                group.data->addElement("name", std::make_shared<adn::StringNavigator>(jointOrWhateverName));

                auto r = convertedAndGroupedData.groups.emplace(groupName, group);
                it = r.first;
            }

            RobotUnitData::RobotUnitDataGroup& group = it->second;
            const std::string escapedName = simox::alg::replace_all(name, ".", "/");
            group.data->addElement(escapedName, RobotUnitDataStreaming::toAron(data.value(), dataEntry));
        }

        auto stop = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
        ARMARX_DEBUG << "RobotUnitData: The total time needed to convert the data to aron is: " << duration;

        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | t Read+Group [ms]", duration.count() / 1000.f);
        }

        return convertedAndGroupedData;
    }


    std::optional<RobotUnitDataStreaming::TimeStep> RobotUnitReader::fetchLatestData()
    {
        std::deque<RobotUnitDataStreaming::TimeStep>& data = receiver->getDataBuffer();
        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | Buffer Size", data.size());
        }
        if (data.empty())
        {
            return std::nullopt;
        }
        else
        {
            RobotUnitDataStreaming::TimeStep currentTimestep = data.back();
            data.clear();
            return currentTimestep;
        }
    }


}

