#pragma once

#include <map>
#include <string>

#include <ArmarXCore/core/logging/Logging.h>

#include "CoreSegmentBase.h"
#include "detail/EntityContainerBase.h"


namespace armarx::armem::base
{

    /**
     * @brief Data of a memory consisting of multiple core segments.
     */
    template <class _CoreSegmentT, class _Derived>
    class MemoryBase :
        public detail::EntityContainerBase<_CoreSegmentT, typename _CoreSegmentT::ProviderSegmentT::EntityT, _Derived>
    {
        using Base = detail::EntityContainerBase<_CoreSegmentT, typename _CoreSegmentT::ProviderSegmentT::EntityT, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using CoreSegmentT = _CoreSegmentT;
        using ProviderSegmentT = typename CoreSegmentT::ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


        struct UpdateResult
        {
            armarx::armem::UpdateType memoryUpdateType;
            armarx::armem::UpdateType coreSegmentUpdateType;
            armarx::armem::UpdateType providerSegmentUpdateType;
            armarx::armem::UpdateType entityUpdateType;
            MemoryID id;
            std::vector<EntitySnapshotT> removedSnapshots;

            UpdateResult() = default;
            UpdateResult(const typename CoreSegmentT::UpdateResult& c) :
                coreSegmentUpdateType(c.coreSegmentUpdateType),
                providerSegmentUpdateType(c.providerSegmentUpdateType),
                entityUpdateType(c.entityUpdateType),
                id(c.id),
                removedSnapshots(c.removedSnapshots)
            {}
        };

    public:

        MemoryBase()
        {
        }
        MemoryBase(const std::string& name) :
            MemoryBase(MemoryID().withMemoryName(name))
        {
        }
        MemoryBase(const MemoryID& id) :
            Base(id)
        {
        }

        MemoryBase(const MemoryBase& other) = default;
        MemoryBase(MemoryBase&& other) = default;
        MemoryBase& operator=(const MemoryBase& other) = default;
        MemoryBase& operator=(MemoryBase&& other) = default;


        inline const std::string& name() const
        {
            return this->id().memoryName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const MemoryBase*>(this)->name());
        }


        inline auto& coreSegments() const
        {
            return this->_container;
        }
        inline auto& coreSegments()
        {
            return this->_container;
        }


        bool hasCoreSegment(const std::string& name) const
        {
            return this->_container.count(name) > 0;
        }

        CoreSegmentT& getCoreSegment(const std::string& name)
        {
            return const_cast<CoreSegmentT&>(const_cast<const MemoryBase*>(this)->getCoreSegment(name));
        }

        const CoreSegmentT& getCoreSegment(const std::string& name) const
        {
            auto it = this->_container.find(name);
            if (it != this->_container.end())
            {
                return it->second;
            }
            else
            {
                throw armem::error::MissingEntry::create<CoreSegmentT>(name, *this);
            }
        }


        bool hasProviderSegment(const MemoryID& providerSegmentID) const
        {
            auto it = this->_container.find(providerSegmentID.coreSegmentName);
            if (it != this->_container.end())
            {
                return it->second.hasProviderSegment(providerSegmentID.providerSegmentName);
            }
            else
            {
                return false;
            }
        }

        ProviderSegmentT& getProviderSegment(const MemoryID& providerSegmentID)
        {
            return getCoreSegment(providerSegmentID.coreSegmentName).getProviderSegment(providerSegmentID.providerSegmentName);
        }

        const ProviderSegmentT& getProviderSegment(const MemoryID& providerSegmentID) const
        {
            return getCoreSegment(providerSegmentID.coreSegmentName).getProviderSegment(providerSegmentID.providerSegmentName);
        }


        using Base::getEntity;
        const EntityT& getEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.memoryName, this->name());
            return getCoreSegment(id.coreSegmentName).getEntity(id);
        }

        const EntityT* findEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.memoryName, this->name());
            if (id.hasCoreSegmentName())
            {
                return getCoreSegment(id.coreSegmentName).findEntity(id);
            }
            else
            {
                for (const auto& [_, coreSegment] : this->_container)
                {
                    if (auto entity = coreSegment.findEntity(id))
                    {
                        return entity;
                    }
                }
                return nullptr;
            }
        }

        /**
         * @brief Add an empty core segment with the given name.
         * @param name The core segment name.
         * @param coreSegmentType The core segment type (optional).
         * @return The added core segment.
         */
        CoreSegmentT& addCoreSegment(const std::string& name, aron::typenavigator::ObjectNavigatorPtr coreSegmentType = nullptr)
        {
            return addCoreSegment(CoreSegmentT(name, coreSegmentType));
        }
        /// Copy and insert a core segment.
        CoreSegmentT& addCoreSegment(const CoreSegmentT& coreSegment)
        {
            return addCoreSegment(CoreSegmentT(coreSegment));
        }
        /// Move and insert a core segment.
        CoreSegmentT& addCoreSegment(CoreSegmentT&& coreSegment)
        {
            if (this->_container.count(coreSegment.name()) > 0)
            {
                throw armem::error::ContainerEntryAlreadyExists(coreSegment.getLevelName(), coreSegment.name(),
                        this->getLevelName(), this->name());
            }
            auto it = this->_container.emplace(coreSegment.name(), std::move(coreSegment)).first;
            it->second.id().setMemoryID(this->id());
            return it->second;
        }

        /**
         * @brief Add multiple core segments.
         * @param The core segment names.
         * @return The core segments. The contained pointers are never null.
         */
        std::vector<CoreSegmentT*> addCoreSegments(const std::vector<std::string>& names)
        {
            std::vector<CoreSegmentT*> segments;
            for (const auto& name : names)
            {
                try
                {
                    segments.push_back(&addCoreSegment(name));
                }
                catch (const armem::error::ContainerEntryAlreadyExists& e)
                {
                    ARMARX_INFO << e.what() << "\nIgnoring multiple addition.";
                }
            }
            return segments;
        }


        /**
         * @brief Store all updates in `commit`.
         * @param commit The commit.
         * @return The resulting memory IDs.
         */
        virtual std::vector<UpdateResult> update(const Commit& commit)
        {
            std::vector<UpdateResult> results;
            for (const auto& update : commit.updates)
            {
                results.push_back(this->update(update));
            }
            return results;
        }

        /**
         * @brief Store the given update.
         * @param update The update.
         * @return The resulting entity snapshot's ID.
         */
        virtual UpdateResult update(const EntityUpdate& update)
        {
            this->_checkContainerName(update.entityID.memoryName, this->name());

            auto [inserted, coreSeg] = addCoreSegmentIfMissing(update.entityID.coreSegmentName);

            // Update entry.
            UpdateResult ret(coreSeg->update(update));
            if (inserted)
            {
                ret.memoryUpdateType = UpdateType::InsertedNew;
            }
            else
            {
                ret.memoryUpdateType = UpdateType::UpdatedExisting;
            }
            return ret;
        }

        std::pair<bool, CoreSegmentT*> addCoreSegmentIfMissing(const std::string& coreSegmentName)
        {
            CoreSegmentT* coreSeg;

            auto it = this->_container.find(coreSegmentName);
            if (it == this->_container.end())
            {
                if (_addMissingCoreSegmentDuringUpdate)
                {
                    // Insert into map.
                    coreSeg = &addCoreSegment(coreSegmentName);
                    return {true, coreSeg};
                }
                else
                {
                    throw error::MissingEntry::create<EntitySnapshotT>(coreSegmentName, *this);
                }
            }
            else
            {
                coreSeg = &it->second;
                return {false, coreSeg};
            }
        }

        /**
         * @brief Merge another memory into this one. Append all data
         * @param m The other memory
         */
        void append(const _Derived& m)
        {
            ARMARX_INFO << "Memory: Merge name '" << m.name() << "' into '" << name() << "'";

            for (const auto& [k, s] : m.container())
            {
                if (const auto& it = this->_container.find(k); it != this->_container.end())
                {
                    // segment already exists
                    it->second.append(s);
                }
                else
                {
                    auto wms = this->_container.emplace(k, this->id().withCoreSegmentName(k));
                    wms.first->second.append(s);
                }
            }
        }

        virtual bool equalsDeep(const MemoryBase& other) const
        {
            //std::cout << "Memory::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, core] : this->_container)
            {
                if (not other.hasCoreSegment(key))
                {
                    return false;
                }
                if (not core.equalsDeep(other.getCoreSegment(key)))
                {
                    return false;
                }
            }
            return true;
        }

        std::string getLevelName() const override
        {
            return "memory";
        }

        std::string getKeyString() const override
        {
            return this->name();
        }

        std::string dump() const
        {
            std::stringstream ss;
            ss << "Memory: " << this->name() << "\n";
            for (const auto& [ckey, cseg] : this->container())
            {
                ss << " |- Found core seg: " << ckey << "\n";
                for (const auto& [pkey, pseg] : cseg.container())
                {
                    ss << " |   |- Found prov seg: " << pkey << "\n";
                    for (const auto& [ekey, eseg] : pseg.container())
                    {
                        ss << " |   |   |- Found entity: " << ekey << "\n";
                    }
                }
            }
            return ss.str();
        }

    protected:

        virtual void _copySelf(DerivedT& other) const override
        {
            Base::_copySelf(other);
            other._addMissingCoreSegmentDuringUpdate = _addMissingCoreSegmentDuringUpdate;
        }
        virtual void _copySelfEmpty(DerivedT& other) const override
        {
            Base::_copySelfEmpty(other);
            other._addMissingCoreSegmentDuringUpdate = _addMissingCoreSegmentDuringUpdate;
        }

    public:

        bool _addMissingCoreSegmentDuringUpdate = false;
    };
}
