#include "Memory.h"

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem::ltm
{

    Memory::Memory(const Memory& other) :
        Base(other),
        dbsettings(other.dbsettings),
        alwaysTransferSettings(other.alwaysTransferSettings),
        periodicTransferSettings(other.periodicTransferSettings),
        onFullTransferSettings(other.onFullTransferSettings),
        mongoDBMutex()
    {
        // Do not copy _mutex.
    }


    Memory::Memory(Memory&& other) :
        Base(other),
        dbsettings(other.dbsettings),
        alwaysTransferSettings(other.alwaysTransferSettings),
        periodicTransferSettings(other.periodicTransferSettings),
        onFullTransferSettings(other.onFullTransferSettings),
        reloaded(other.reloaded)
    {
        // Do not move _mutex.
    }


    Memory& Memory::operator=(const Memory& other)
    {
        Base::operator=(other);
        dbsettings = other.dbsettings;
        alwaysTransferSettings = other.alwaysTransferSettings;
        periodicTransferSettings = other.periodicTransferSettings;
        onFullTransferSettings = other.onFullTransferSettings;

        // Don't copy _mutex.
        return *this;
    }


    Memory& Memory::operator=(Memory&& other)
    {
        Base::operator=(other);
        dbsettings = std::move(other.dbsettings);
        alwaysTransferSettings = std::move(other.alwaysTransferSettings);
        periodicTransferSettings = std::move(other.periodicTransferSettings);
        onFullTransferSettings = std::move(other.onFullTransferSettings);
        reloaded = other.reloaded;

        // Don't move _mutex.
        return *this;
    }

    bool Memory::checkConnection() const
    {
        // Check connection:
        ARMARX_INFO << "Checking connection";
        if (!MongoDBConnectionManager::ConnectionIsValid(dbsettings))
        {
            ARMARX_WARNING << deactivateSpam("ConnectionIsNotValid")
                           << "The connection to mongocxx for memory '" << name() << "' is not valid. Settings are: " << dbsettings.toString()
                           << "\nTo start it, run e.g.: \n"
                           << "mongod --port " << dbsettings.port << " --dbpath \"/tmp/\""
                           << "\n\n";
            return false;
        }
        return true;
    }

    void Memory::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(dbsettings.host, prefix + ".ltm.10_host");
        defs->optional(dbsettings.port, prefix + "ltm.11_port");
        defs->optional(dbsettings.user, prefix + "ltm.20_user");
        defs->optional(dbsettings.password, prefix + "ltm.21_password");
        defs->optional(dbsettings.database, prefix + "ltm.22_database");
        defs->optional(alwaysTransferSettings.enabled, prefix + "ltm.30_enableAlwaysTransfer", "Enable transfer whenever new data is committed (This disables the other transfer modes!).");
        defs->optional(periodicTransferSettings.enabled, prefix + "ltm.31_enablePeriodicTransfer", "Enable transfer based on periodic interval.");
        defs->optional(onFullTransferSettings.enabled, prefix + "ltm.32_enableOnFullTransfer", "Enable transfer whenever the wm is full (see maxHistorySize).");
    }

    wm::Memory Memory::convert() const
    {
        std::lock_guard l(mongoDBMutex);
        if (!checkConnection())
        {
            wm::Memory m(id());
            return m;
        }

        ARMARX_INFO << "Converting Memory with connection to: " << dbsettings.toString();

        TIMING_START(LTM_Convert);

        wm::Memory m(id());
        for (const auto& [_, s] : _container)
        {
            m.addCoreSegment(s.convert());
        }

        TIMING_END(LTM_Convert);
        return m;
    }

    void Memory::reload()
    {
        std::lock_guard l(mongoDBMutex);
        reloaded = false;

        if (!checkConnection())
        {
            return;
        }

        ARMARX_INFO << "(Re)Establishing connection to: " << dbsettings.toString();

        TIMING_START(LTM_Reload);
        _container.clear();

        mongocxx::client& client = MongoDBConnectionManager::EstablishConnection(dbsettings);
        if (!client)
        {
            ARMARX_ERROR << "A client has died. Could not reload.";
            return;
        }

        auto databases = client.list_databases();
        for (const auto& doc : databases)
        {
            nlohmann::json json = nlohmann::json::parse(bsoncxx::to_json(doc));
            ARMARX_INFO << "Found the database: " << json.at("name");
        }

        mongocxx::database db = client[dbsettings.database];
        ARMARX_INFO << "Getting collection for id: " << id().str();
        mongocxx::collection coll = db[id().str()];

        ARMARX_IMPORTANT << "Memory Container size is: " << _container.size();

        mongocxx::cursor cursor = coll.find({});
        for (const auto& doc : cursor)
        {
            nlohmann::json json = nlohmann::json::parse(bsoncxx::to_json(doc));
            ARMARX_INFO << "Memory: Found foreign key: " << json.at("foreign_key");

            MemoryID i((std::string) json.at("foreign_key"));
            if (i.memoryName != id().memoryName)
            {
                throw error::InvalidMemoryID(i, "A MemoryID in mongodb was invalid. Found the wrong memory name. Expected " + id().memoryName);
            }

            std::string k = i.coreSegmentName;

            if (const auto& it = _container.find(k); it != _container.end())
            {
                throw error::ArMemError("Somehow after clearing the (memory) container a key k = " + k + " was found. Do you have double entries in mongodb?");
            }
            else
            {
                auto wms = _container.emplace(k, id().withCoreSegmentName(k));
                wms.first->second.dbsettings = dbsettings;
                wms.first->second.reload();
            }
        }

        reloaded = true;
        for (const auto& m : toAppendQueue)
        {
            _append(m);
        }

        TIMING_END(LTM_Reload);
        ARMARX_INFO << "After reload memory " << id().str() << " size: " << _container.size() << ". Setting reloaded: " << reloaded;
    }

    void Memory::_append(const wm::Memory& m)
    {
        if (!checkConnection() || !reloaded)
        {
            // We ignore if not fully loaded yet
            return;
        }

        //ARMARX_INFO << "Merge memory with name '" << m.name() << "' into the LTM with name '" << name() << "'";

        TIMING_START(LTM_Append);

        mongocxx::client& client = MongoDBConnectionManager::EstablishConnection(dbsettings);
        mongocxx::database db = client[dbsettings.database];
        mongocxx::collection coll = db[id().str()];

        for (const auto& [k, s] : m.container())
        {
            if (const auto& it = _container.find(k); it != _container.end())
            {
                // TODO check if foreign key exists
                it->second.append(s);
            }
            else
            {
                auto builder = bsoncxx::builder::stream::document{};
                bsoncxx::document::value foreign_key = builder
                                                       << "foreign_key" << s.id().withCoreSegmentName(k).str()
                                                       << bsoncxx::builder::stream::finalize;
                coll.insert_one(foreign_key.view());

                auto wms = _container.emplace(k, id().withCoreSegmentName(k));
                wms.first->second.dbsettings = dbsettings;
                wms.first->second.append(s);
            }
        }

        TIMING_END(LTM_Append);
    }

    void Memory::append(const wm::Memory& m)
    {
        std::lock_guard l(mongoDBMutex);
        if (!checkConnection() || !reloaded)
        {
            toAppendQueue.push_back(m);
            return;
        }

        _append(m);
    }
}
