#include "CoreSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem::wm
{

    CoreSegment::CoreSegment(const CoreSegment& other) :
        CoreSegment::Base(other),
        _mutex()
    {
        // Do not copy _mutex.
    }


    CoreSegment::CoreSegment(CoreSegment&& other) : CoreSegment::Base(other)
    {
        // Do not move _mutex.
    }


    CoreSegment& CoreSegment::operator=(const CoreSegment& other)
    {
        Base::operator=(other);
        // Don't copy _mutex.
        return *this;
    }


    CoreSegment& CoreSegment::operator=(CoreSegment&& other)
    {
        Base::operator=(other);
        // Don't move _mutex.
        return *this;
    }


    Commit CoreSegment::toCommit() const
    {
        Commit c;
        for (const auto& [k, s] : _container)
        {
            c.append(s.toCommit());
        }
        return c;
    }


    void CoreSegment::_copySelfWithoutData(CoreSegment& other) const
    {
        other.id() = _id;
        other.setMaxHistorySize(_maxHistorySize);
        for (const auto& [k, s] : _container)
        {
            other.addProviderSegment(s.copyWithoutData());
        }
    }


    std::mutex& CoreSegment::mutex() const
    {
        return _mutex;
    }


    std::optional<wm::EntitySnapshot> CoreSegment::getLatestEntitySnapshot(const MemoryID& entityID) const
    {
        const wm::Entity& entity = this->getEntity(entityID);
        if (entity.empty())
        {
            return std::nullopt;
        }
        else
        {
            return entity.getLatestSnapshot();
        }
    }


    std::optional<wm::EntityInstance> CoreSegment::getLatestEntityInstance(
        const MemoryID& entityID, int instanceIndex) const
    {
        auto snapshot = getLatestEntitySnapshot(entityID);
        if (snapshot.has_value()
            and instanceIndex >= 0
            and static_cast<size_t>(instanceIndex) < snapshot->size())
        {
            return snapshot->getInstance(instanceIndex);
        }
        else
        {
            return std::nullopt;
        }
    }


    armarx::aron::datanavigator::DictNavigatorPtr
    CoreSegment::getLatestEntityInstanceData(const MemoryID& entityID, int instanceIndex) const
    {
        auto instance = getLatestEntityInstance(entityID, instanceIndex);
        if (instance.has_value())
        {
            return instance->data();
        }
        else
        {
            return nullptr;
        }
    }


    std::optional<wm::EntitySnapshot>
    CoreSegment::getLatestEntitySnapshotLocking(const MemoryID& entityID) const
    {
        std::scoped_lock lock(_mutex);
        return getLatestEntitySnapshot(entityID);
    }

    std::optional<wm::EntityInstance>
    CoreSegment::getLatestEntityInstanceLocking(const MemoryID& entityID, int instanceIndex) const
    {
        std::scoped_lock lock(_mutex);
        return getLatestEntityInstance(entityID, instanceIndex);
    }

    armarx::aron::datanavigator::DictNavigatorPtr
    CoreSegment::getLatestEntityInstanceDataLocking(const MemoryID& entityID, int instanceIndex) const
    {
        std::scoped_lock lock(_mutex);
        return getLatestEntityInstanceData(entityID, instanceIndex);
    }

}
