#pragma once

#include <mutex>
#include <optional>

#include "../base/CoreSegmentBase.h"

#include "ProviderSegment.h"
#include "detail/CopyWithoutData.h"


namespace armarx::armem::wm
{

    /**
     * @brief Data of a core segment containing multiple provider segments.
     */
    class CoreSegment :
        public base::CoreSegmentBase<ProviderSegment, CoreSegment>,
        public detail::CopyWithoutData<CoreSegment>
    {
        using Base = base::CoreSegmentBase<ProviderSegment, CoreSegment>;

    public:

        using Base::CoreSegmentBase;

        CoreSegment(const CoreSegment& other);
        CoreSegment(CoreSegment&& other);
        CoreSegment& operator=(const CoreSegment& other);
        CoreSegment& operator=(CoreSegment&& other);

        /**
         * @brief Convert the content of this segmnet into a commit
         * @return The resulting commit
         */
        Commit toCommit() const;

        std::mutex& mutex() const;


        // Non-locking interface

        std::optional<wm::EntitySnapshot> getLatestEntitySnapshot(
            const MemoryID& entityID) const;
        std::optional<wm::EntityInstance> getLatestEntityInstance(
            const MemoryID& entityID, int instanceIndex = 0) const;
        armarx::aron::datanavigator::DictNavigatorPtr getLatestEntityInstanceData(
            const MemoryID& entityID, int instanceIndex = 0) const;


        template <class AronDtoT>
        std::optional<AronDtoT> getLatestEntityInstanceDataAs(
            const MemoryID& entityID, int instanceIndex = 0) const
        {
            wm::EntityInstanceDataPtr data = getLatestEntityInstanceData(entityID, instanceIndex);
            if (data)
            {
                AronDtoT aron;
                aron.fromAron(data);
                return aron;
            }
            else
            {
                return std::nullopt;
            }
        }


        // Locking interface

        std::optional<wm::EntitySnapshot> getLatestEntitySnapshotLocking(
            const MemoryID& entityID) const;
        std::optional<wm::EntityInstance> getLatestEntityInstanceLocking(
            const MemoryID& entityID, int instanceIndex = 0) const;
        armarx::aron::datanavigator::DictNavigatorPtr getLatestEntityInstanceDataLocking(
            const MemoryID& entityID, int instanceIndex = 0) const;


        template <class AronDtoT>
        std::optional<AronDtoT> getLatestEntityInstanceDataLockingAs(
            const MemoryID& entityID, int instanceIndex = 0) const
        {
            // Keep lock to a minimum.
            wm::EntityInstanceDataPtr data = nullptr;
            {
                std::scoped_lock lock(_mutex);
                data = getLatestEntityInstanceData(entityID, instanceIndex);
            }
            if (data)
            {
                AronDtoT aron;
                aron.fromAron(data);
                return aron;
            }
            else
            {
                return std::nullopt;
            }
        }


    protected:

        virtual void _copySelfWithoutData(CoreSegment& other) const override;

        mutable std::mutex _mutex;

    };

}
