#include "MemoryToIceAdapter.h"

#include "ArmarXCore/core/logging/Logging.h"
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"
#include "../core/workingmemory/ice_conversions.h"
#include "query_proc/workingmemory/MemoryQueryProcessor.h"
#include "query_proc/longtermmemory/MemoryQueryProcessor.h"


namespace armarx::armem::server
{

    MemoryToIceAdapter::MemoryToIceAdapter(wm::Memory* workingMemory, ltm::Memory* longtermMemory) :
        workingMemory(workingMemory), longtermMemory(longtermMemory)
    {
    }


    void
    MemoryToIceAdapter::setMemoryListener(client::MemoryListenerInterfacePrx memoryListener)
    {
        this->memoryListenerTopic = memoryListener;
    }


    // WRITING
    data::AddSegmentResult
    MemoryToIceAdapter::addSegment(const data::AddSegmentInput& input, bool addCoreSegments)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        data::AddSegmentResult output;

        armem::wm::CoreSegment* coreSegment = nullptr;
        try
        {
            coreSegment = &workingMemory->getCoreSegment(input.coreSegmentName);
        }
        catch (const armem::error::MissingEntry& e)
        {
            if (addCoreSegments)
            {
                coreSegment = &workingMemory->addCoreSegment(input.coreSegmentName);
            }
            else
            {
                output.success = false;
                output.errorMessage = e.what();
                return output;
            }
        }
        ARMARX_CHECK_NOT_NULL(coreSegment);

        if (input.providerSegmentName.size() > 0)
        {
            std::scoped_lock lock(coreSegment->mutex());
            try
            {
                coreSegment->addProviderSegment(input.providerSegmentName);
            }
            catch (const armem::error::ContainerEntryAlreadyExists&)
            {
                // This is ok.
                if (input.clearWhenExists)
                {
                    wm::ProviderSegment& provider = coreSegment->getProviderSegment(input.providerSegmentName);
                    provider.clear();
                }
            }
        }

        armem::MemoryID segmentID;
        segmentID.memoryName = workingMemory->name();
        segmentID.coreSegmentName = input.coreSegmentName;
        segmentID.providerSegmentName = input.providerSegmentName;

        output.success = true;
        output.segmentID = segmentID.str();
        return output;
    }


    data::AddSegmentsResult
    MemoryToIceAdapter::addSegments(const data::AddSegmentsInput& input, bool addCoreSegments)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        data::AddSegmentsResult output;
        for (const auto& i : input)
        {
            output.push_back(addSegment(i, addCoreSegments));
        }
        return output;
    }


    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce, Time timeArrived)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);
        auto handleException = [](const std::string & what)
        {
            data::CommitResult result;
            data::EntityUpdateResult& r = result.results.emplace_back();
            r.success = false;
            r.errorMessage = what;
            return result;
        };

        armem::Commit commit;
        try
        {
            armem::fromIce(commitIce, commit, timeArrived);
        }
        catch (const aron::error::AronNotValidException& e)
        {
            throw;
            return handleException(e.what());
        }
        catch (const Ice::Exception& e)
        {
            throw;
            return handleException(e.what());
        }

        armem::CommitResult result = this->commit(commit);
        data::CommitResult resultIce;
        armem::toIce(resultIce, result);

        return resultIce;
    }


    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce)
    {
        return commit(commitIce, armem::Time::now());
    }


    armem::CommitResult
    MemoryToIceAdapter::commit(const armem::Commit& commit)
    {
        return this->_commit(commit, false);
    }


    armem::CommitResult MemoryToIceAdapter::commitLocking(const armem::Commit& commit)
    {
        return this->_commit(commit, true);
    }


    armem::CommitResult MemoryToIceAdapter::_commit(const armem::Commit& commit, bool locking)
    {
        std::vector<data::MemoryID> updatedIDs;
        const bool publishUpdates = bool(memoryListenerTopic);

        CommitResult commitResult;
        for (const EntityUpdate& update : commit.updates)
        {
            EntityUpdateResult& result = commitResult.results.emplace_back();
            try
            {
                auto updateResult = locking
                                    ? workingMemory->updateLocking(update)
                                    : workingMemory->update(update);

                result.success = true;
                result.snapshotID = updateResult.id;
                result.timeArrived = update.timeArrived;

                // also store in ltm if transfermode is set to always
                // TODO: Move outside of loop?
                if (longtermMemory)
                {
                    if (longtermMemory->alwaysTransferSettings.enabled)
                    {
                        wm::Memory tmp(longtermMemory->name());
                        tmp._addMissingCoreSegmentDuringUpdate = true;
                        tmp.update(update);
                        longtermMemory->append(tmp);
                    }
                    else
                    {
                        // TODO: see below
                    }
                }

                // TODO: Consollidate to ltm if onFilledTransfer is enabled (fabian.peller)
                for (const auto& snapshot : updateResult.removedSnapshots)
                {
                    ARMARX_DEBUG << "The id " << snapshot.id() << " was removed from wm";
                }

                if (publishUpdates)
                {
                    data::MemoryID& id = updatedIDs.emplace_back();
                    armem::toIce(id, result.snapshotID);
                }
            }
            catch (const error::ArMemError& e)
            {
                result.success = false;
                result.errorMessage = e.what();
            }
            catch (const aron::error::AronException& e)
            {
                result.success = false;
                result.errorMessage = e.what();
            }
            catch (const Ice::Exception& e)
            {
                result.success = false;
                result.errorMessage = e.what();
            }
        }

        if (publishUpdates)
        {
            memoryListenerTopic->memoryUpdated(updatedIDs);
        }

        return commitResult;
    }


    // READING
    armem::query::data::Result
    MemoryToIceAdapter::query(const armem::query::data::Input& input)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        // Core segment processors will aquire the core segment locks.
        armem::wm::query_proc::MemoryQueryProcessor wmProcessor(
            input.withData ? armem::DataMode::WithData : armem::DataMode::NoData);
        wm::Memory wmResult = wmProcessor.process(input, *workingMemory, /* execute if: */ { query::data::QueryTarget::WM });

        armem::ltm::query_proc::MemoryQueryProcessor ltmProcessor;
        ltm::Memory ltmResult = ltmProcessor.process(input, *longtermMemory, /* execute if: */ { query::data::QueryTarget::LTM });

        armem::query::data::Result result;
        if (ltmResult.hasData())
        {
            ARMARX_INFO << "The LTM returned data after query";

            // ATTENTION: This code block moves data from LTM back into WM.
            // However, since some segments are constrained, the WM might send data back to LTM.
            // This may also affect the data returned by the current query.
            // However, this is expected behavior, since we copy the data in the processor (copyEmpty) we can safely return the copy and
            // remove the original memory reference from WM here.
            wm::Memory ltmConverted = ltmResult.convert();
            if (!ltmConverted.hasData())
            {
                ARMARX_ERROR << "A converted memory contains no data although the original memory contained data. This indicates that something is wrong.";
            }

            wmResult.append(ltmConverted);
            if (!wmResult.hasData())
            {
                ARMARX_ERROR << "A merged Memory has no data although at least the LTM result contains data. This indicates that something is wrong.";
            }

            // query again to limit output size (TODO: Skip if querytype is all)
            auto queryInput = armem::client::QueryInput::fromIce(input);
            queryInput.replaceQueryTarget(query::data::QueryTarget::LTM, query::data::QueryTarget::WM);

            wm::Memory merged_result = wmProcessor.process(queryInput.toIce(), wmResult, /* execute if: */ { query::data::QueryTarget::WM });
            if (!merged_result.hasData())
            {
                ARMARX_ERROR << "A merged and postprocessed Memory has no data although at least the LTM result contains data. This indicates that something is wrong.";
            }

            result.memory = toIce<data::MemoryPtr>(merged_result);

            // also move results of ltm to wm
            //this->commit(ltm_converted.toCommit());

            // mark removed entries of wm in viewer
            // TODO
        }
        else
        {
            ARMARX_INFO << "The LTM did not return data after query";
            result.memory = toIce<data::MemoryPtr>(wmResult);
        }

        result.success = true;
        if (result.memory->coreSegments.size() == 0)
        {
            ARMARX_DEBUG << "No data in memory found after query.";
        }

        return result;
    }


    client::QueryResult MemoryToIceAdapter::query(const client::QueryInput& input)
    {
        return client::QueryResult::fromIce(query(input.toIce()));
    }


    // LTM LOADING FROM LTM
    query::data::Result MemoryToIceAdapter::load(const armem::query::data::Input& query)
    {
        ARMARX_CHECK_NOT_NULL(longtermMemory);
        query::data::Result output;

        output.success = true;
        return output;
    }


    // LTM STORING
    data::StoreResult MemoryToIceAdapter::store(const armem::data::StoreInput& input)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);
        ARMARX_CHECK_NOT_NULL(longtermMemory);
        data::StoreResult output;

        for (const auto& query : input.query.memoryQueries)
        {
            // force query to only query WM
            // TODO: Think about other query class (fabian.peller)
            query->targets = {query::data::QueryTarget::WM};
        }

        armem::query::data::Result queryResult = this->query(input.query);
        output.success = queryResult.success;

        if (queryResult.success)
        {
            wm::Memory m;
            fromIce(queryResult.memory, m);
            longtermMemory->append(m);
        }

        return output;
    }





}
