#pragma once

// STD/STL
#include <mutex>
#include <string>

// ArmarX
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/application/properties/PluginAll.h>
#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>

#include "../MemoryToIceAdapter.h"


namespace armarx::armem::server::segment
{
    namespace detail
    {
        /**
         * @brief A base class for memory servers to manage their segments.
         * A segment can inherit this base class and add segment specific code.
         * TODO (fabian.peller): add concept to only accept coresegments, providersegments or entitysegments
         */
        template <class SegmentType>
        class SegmentBase : public armarx::Logging
        {
        public:
            SegmentBase() = delete;
            SegmentBase(armem::server::MemoryToIceAdapter& iceMemory) :
                iceMemory(iceMemory)
            {
                Logging::setTag("armarx::armem::Segment");
            }

            virtual ~SegmentBase() = default;

            virtual void defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix = "") = 0;
            virtual void onInit() = 0;
            virtual void onConnect() = 0;

        protected:
            // Memory connection
            armem::server::MemoryToIceAdapter& iceMemory;

            SegmentType* segment;

        private:
        };
    }

    namespace wm
    {
        /**
         * @brief A base class for core segments
         */
        class CoreSegmentBase : public detail::SegmentBase<armarx::armem::wm::CoreSegment>
        {
            using Base = detail::SegmentBase<armarx::armem::wm::CoreSegment>;

        public:
            CoreSegmentBase(armem::server::MemoryToIceAdapter& iceMemory,
                            const std::string& defaultCoreSegmentName = "",
                            aron::typenavigator::ObjectNavigatorPtr coreSegmentAronType = nullptr,
                            int defaultMaxHistorySize = -1):
                Base(iceMemory),
                p({defaultCoreSegmentName, defaultMaxHistorySize}),
            coreSegmentAronType(coreSegmentAronType)
            {
                Logging::setTag("armarx::armem::wm::Segment");
            }

            virtual ~CoreSegmentBase() = default;

            virtual void defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix = "") override
            {
                ARMARX_CHECK_NOT_NULL(defs);

                defs->optional(p.coreSegmentName, prefix + "CoreSegmentName", "Name of the '" + p.coreSegmentName + "' core segment.");
                defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of '" + p.coreSegmentName + "' history (-1 for infinite).");
            }

            virtual void onInit() override
            {
                ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

                segment = &iceMemory.workingMemory->addCoreSegment(p.coreSegmentName, coreSegmentAronType);
                segment->setMaxHistorySize(p.maxHistorySize);
            }

            virtual void onConnect() override
            {

            }

            std::mutex& mutex() const
            {
                ARMARX_CHECK_NOT_NULL(segment);
                return segment->mutex();
            }

        protected:
            struct Properties
            {
                std::string coreSegmentName;
                int maxHistorySize;
            };
            Properties p;

            aron::typenavigator::ObjectNavigatorPtr coreSegmentAronType;
        };


        /**
         * @brief A base class for provider segments
         */
        class ProviderSegmentBase : public detail::SegmentBase<armarx::armem::wm::ProviderSegment>
        {
            using Base = detail::SegmentBase<armarx::armem::wm::ProviderSegment>;

        public:
            ProviderSegmentBase(armem::server::MemoryToIceAdapter& iceMemory,
                                const std::string& defaultProviderSegmentName = "",
                                aron::typenavigator::ObjectNavigatorPtr providerSegmentAronType = nullptr,
                                const std::string& defaultCoreSegmentName = "",
                                aron::typenavigator::ObjectNavigatorPtr coreSegmentAronType = nullptr,
                                int defaultMaxHistorySize = -1):
                Base(iceMemory),
                p({defaultCoreSegmentName, defaultProviderSegmentName, defaultMaxHistorySize}),
            coreSegmentAronType(coreSegmentAronType),
            providerSegmentAronType(providerSegmentAronType)
            {
                Logging::setTag("armarx::armem::wm::Segment");
            }

            virtual ~ProviderSegmentBase() = default;

            virtual void defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix = "") override
            {
                ARMARX_CHECK_NOT_NULL(defs);

                defs->optional(p.coreSegmentName, prefix + "CoreSegmentName", "Name of the '" + p.coreSegmentName + "' core segment.");
                defs->optional(p.providerSegmentName, prefix + "ProviderSegmentName", "Name of the '" + p.providerSegmentName + "' provider segment.");
                defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of '" + p.providerSegmentName + "' history (-1 for infinite).");
            }

            virtual void onInit() override
            {
                ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

                if (!iceMemory.workingMemory->hasCoreSegment(p.coreSegmentName))
                {
                    coreSegment = &iceMemory.workingMemory->addCoreSegment(p.coreSegmentName, coreSegmentAronType);
                    coreSegment->setMaxHistorySize(p.maxHistorySize);
                }
                else
                {
                    coreSegment = &iceMemory.workingMemory->getCoreSegment(p.coreSegmentName);
                }

                segment = &coreSegment->addProviderSegment(p.providerSegmentName, providerSegmentAronType);
                segment->setMaxHistorySize(p.maxHistorySize);
            }

            virtual void onConnect() override
            {

            }

        protected:
            struct Properties
            {
                std::string coreSegmentName;
                std::string providerSegmentName;
                int maxHistorySize;
            };
            Properties p;

            aron::typenavigator::ObjectNavigatorPtr coreSegmentAronType;
            aron::typenavigator::ObjectNavigatorPtr providerSegmentAronType;

            armarx::armem::wm::CoreSegment* coreSegment;
        };


        /**
         * @brief A base class for provider segments with an aron business object
         */
        template <class BusinessClassObject>
        class AronTypedCoreSegmentBase : public CoreSegmentBase
        {
            using Base = CoreSegmentBase;

        public:
            AronTypedCoreSegmentBase(armem::server::MemoryToIceAdapter& iceMemory,
                                     const std::string& defaultCoreSegmentName = "",
                                     int defaultMaxHistorySize = -1):
                Base(iceMemory, defaultCoreSegmentName, BusinessClassObject::toAronType(), defaultMaxHistorySize)
            {
            }

            virtual ~AronTypedCoreSegmentBase() = default;
        };


        /**
         * @brief A base class for provider segments with an aron business object
         */
        template <class BusinessClassObject>
        class AronTypedProviderSegmentBase : public ProviderSegmentBase
        {
            using Base = ProviderSegmentBase;

        public:
            AronTypedProviderSegmentBase(armem::server::MemoryToIceAdapter& iceMemory,
                                         const std::string& defaultProviderSegmentName = "",
                                         const std::string& defaultCoreSegmentName = "",
                                         int defaultMaxHistorySize = -1):
                Base(iceMemory, defaultProviderSegmentName, BusinessClassObject::toAronType(), defaultCoreSegmentName, nullptr, defaultMaxHistorySize)
            {
            }

            virtual ~AronTypedProviderSegmentBase() = default;
        };
    }
}
