#include "Visu.h"

#include <algorithm>
#include <exception>
#include <string>

#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/math/pose.h>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <RobotAPI/components/ArViz/Client/Elements.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>

#include "../description/Segment.h"
#include "../localization/Segment.h"
#include "../proprioception/Segment.h"


namespace armarx::armem::server::robot_state
{

    Visu::Visu(const description::Segment& descriptionSegment, const proprioception::Segment& proprioceptionSegment, const localization::Segment& localizationSegment)
        : descriptionSegment(descriptionSegment),
          proprioceptionSegment(proprioceptionSegment),
          localizationSegment(localizationSegment)
    {}


    void Visu::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(
            p.enabled, prefix + "enabled", "Enable or disable visualization of objects.");
        defs->optional(p.frequencyHz, prefix + "frequenzyHz", "Frequency of visualization.");
    }


    void Visu::init()
    {
    }


    void Visu::connect(const viz::Client& arviz, DebugObserverInterfacePrx debugObserver)
    {
        this->arviz = arviz;
        if (debugObserver)
        {
            bool batchMode = true;
            this->debugObserver = DebugObserverHelper("RobotStateMemory", debugObserver, batchMode);
        }

        updateTask = new SimpleRunningTask<>([this]()
        {
            this->visualizeRun();
        });
        updateTask->start();
    }


    void Visu::visualizeRobots(viz::Layer& layer, const robot::Robots& robots)
    {
        const auto visualizeRobot = [&](const robot::Robot & robot)
        {
            const auto xmlPath = robot.description.xml.serialize();

            // clang-format off
            auto robotVisu = viz::Robot(robot.description.name)
                             //  .file(xmlPath.package, xmlPath.path)
                             .file(xmlPath.package, xmlPath.path)
                             .joints(robot.config.jointMap)
                             .pose(robot.config.globalPose);

            robotVisu.useFullModel();
            // clang-format on

            layer.add(robotVisu);
        };

        std::for_each(robots.begin(), robots.end(), visualizeRobot);
    }


    void Visu::visualizeFrames(
        viz::Layer& layerFrames,
        const std::unordered_map<std::string, std::vector<Eigen::Affine3f>>& frames)
    {
        for (const auto& [robotName, robotFrames] : frames)
        {
            Eigen::Affine3f pose = Eigen::Affine3f::Identity();

            int i = 0;
            for (const auto& frame : robotFrames)
            {
                Eigen::Affine3f from = pose;
                Eigen::Affine3f to = pose * frame;

                layerFrames.add(viz::Arrow(robotName + std::to_string(++i)).fromTo(from.translation(), to.translation()));
            }
        }
    }


    robot::Robots combine(
        const std::unordered_map<std::string, robot::RobotDescription>& robotDescriptions,
        const std::unordered_map<std::string, Eigen::Affine3f>& globalRobotPoseMap,
        const std::unordered_map<std::string, std::map<std::string, float>>& robotJointPositionMap)
    {

        robot::Robots robots;

        for (const auto &[robotName, robotDescription] : robotDescriptions)
        {
            const auto& globalPose = globalRobotPoseMap.find(robotName);
            if (globalPose == globalRobotPoseMap.end())
            {
                ARMARX_WARNING << deactivateSpam(10) << "No global pose for robot '" << robotName
                               << "'";
                continue;
            }

            const auto& jointMap = robotJointPositionMap.find(robotName);
            if (jointMap == robotJointPositionMap.end())
            {
                ARMARX_WARNING << deactivateSpam(10) << "No joint positions for robot '"
                               << robotName << "'";
                continue;
            }

            ARMARX_DEBUG << "Found the following joints for robot " << robotName << ": "
                         << simox::alg::get_keys(jointMap->second);

            // TODO(fabian.reister): based on data
            const armem::Time timestamp = IceUtil::Time::now();

            robots.emplace_back(robot::Robot
            {
                .description = robotDescription,
                .instance    = "", // TODO(fabian.reister): set this properly
                .config      = {
                    .timestamp  = timestamp,
                    .globalPose = globalPose->second,
                    .jointMap   = jointMap->second
                },
                .timestamp   = timestamp,
            });
        }

        return robots;
    }


    void Visu::visualizeRun()
    {
        CycleUtil cycle(static_cast<int>(1000 / p.frequencyHz));
        while (updateTask && not updateTask->isStopped())
        {
            {
                // std::scoped_lock lock(visuMutex);
                ARMARX_DEBUG << "Update task";
                if (p.enabled)
                {
                    TIMING_START(tVisuTotal);

                    // TODO(fabian.reister): use timestamp
                    const Time timestamp = Time::now();

                    try
                    {
                        // Get data.
                        TIMING_START(tVisuGetData);

                        TIMING_START(tRobotDescriptions);
                        const description::Segment::RobotDescriptionMap robotDescriptions =
                            descriptionSegment.getRobotDescriptionsLocking(timestamp);
                        TIMING_END_STREAM(tRobotDescriptions, ARMARX_DEBUG);

                        TIMING_START(tGlobalRobotPoseMap);
                        const auto globalRobotPoseMap =
                            localizationSegment.getRobotGlobalPosesLocking(timestamp);
                        TIMING_END_STREAM(tGlobalRobotPoseMap, ARMARX_DEBUG);

                        TIMING_START(tRobotFramePoses);
                        const auto frames = localizationSegment.getRobotFramePosesLocking(timestamp);
                        TIMING_END_STREAM(tRobotFramePoses, ARMARX_DEBUG);

                        TIMING_START(tRobotJointPositionMap);
                        const auto robotJointPositionMap =
                            proprioceptionSegment.getRobotJointPositionsLocking(
                                timestamp, debugObserver ? &*debugObserver : nullptr);
                        TIMING_END_STREAM(tRobotJointPositionMap, ARMARX_DEBUG);

                        TIMING_END_STREAM(tVisuGetData, ARMARX_DEBUG);


                        // Build layers.
                        TIMING_START(tVisuBuildLayers);

                        // we need all 3 informations:
                        // - robot description
                        // - global pose
                        // - joint positions
                        // => this is nothing but a armem::Robot

                        const robot::Robots robots =
                            combine(robotDescriptions, globalRobotPoseMap, robotJointPositionMap);

                        ARMARX_DEBUG << "Visualize robots ...";
                        viz::Layer layer = arviz.layer("Robots");
                        visualizeRobots(layer, robots);

                        ARMARX_DEBUG << "Visualize frames ...";
                        viz::Layer layerFrames = arviz.layer("Frames");
                        visualizeFrames(layerFrames, frames);

                        TIMING_END_STREAM(tVisuBuildLayers, ARMARX_DEBUG);


                        // Commit layers.

                        ARMARX_DEBUG << "Commit visualization ...";
                        TIMING_START(tVisuCommit);
                        arviz.commit({layer, layerFrames});
                        TIMING_END_STREAM(tVisuCommit, ARMARX_DEBUG);

                        TIMING_END_STREAM(tVisuTotal, ARMARX_DEBUG);

                        if (debugObserver.has_value())
                        {
                            const std::string p = "Visu | ";
                            debugObserver->setDebugObserverDatafield(p + "t Total (ms)", tVisuTotal.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 1 Get Data (ms)", tVisuGetData.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 1.1 Descriptions (ms)", tRobotDescriptions.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 1.2 Global Pose (ms)", tGlobalRobotPoseMap.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 1.3 Frames (ms)", tRobotFramePoses.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 1.4 Joint Positions (ms)", tRobotJointPositionMap.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 2 Build Layers (ms)", tVisuBuildLayers.toMilliSecondsDouble());
                            debugObserver->setDebugObserverDatafield(p + "t 3 Commit (ms)", tVisuCommit.toMilliSecondsDouble());
                        }
                    }
                    catch (const std::exception& ex)
                    {
                        ARMARX_WARNING << "Caught exception while visualizing robots: \n" << ex.what();
                        continue;
                    }
                    catch (...)
                    {
                        ARMARX_WARNING << "Caught unknown exception while visualizing robots.";
                        continue;
                    }

                    if (debugObserver.has_value())
                    {
                        debugObserver->sendDebugObserverBatch();
                    }
                }
            }
            cycle.waitForCycleDuration();
        }
    }

} // namespace armarx::armem::server::robot_state
