#include "Segment.h"

// STL
#include <iterator>
#include <sstream>

// Ice
#include <IceUtil/Time.h>

// Eigen
#include <Eigen/Core>
#include <Eigen/Geometry>

#include <ArmarXCore/core/application/properties/PropertyDefinitionContainer.h>
#include <ArmarXCore/core/application/properties/PluginAll.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <RobotAPI/libraries/core/FramedPose.h>

#include <RobotAPI/libraries/aron/common/aron_conversions.h>

#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/workingmemory/visitor.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>

#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/robot_conversions.h>

#include <RobotAPI/libraries/armem_robot_state/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/TransformHelper.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/types.h>


namespace armarx::armem::server::robot_state::localization
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        Base(memoryToIceAdapter, "Localization", 1024)
    {
    }


    Segment::~Segment() = default;


    Segment::RobotFramePoseMap
    Segment::getRobotFramePosesLocking(const armem::Time& timestamp) const
    {
        std::scoped_lock lock(mutex());
        return getRobotFramePoses(timestamp);
    }


    Segment::RobotFramePoseMap
    Segment::getRobotFramePoses(const armem::Time& timestamp) const
    {
        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        RobotFramePoseMap frames;

        const auto knownRobots = simox::alg::get_keys(*segment);
        for (const auto& robotName : knownRobots)
        {
            TransformQuery query{.header{.parentFrame = GlobalFrame,
                                         .frame       = "root", // TODO, FIXME
                                         .agent       = robotName,
                                         .timestamp   = timestamp}};

            const auto result = TransformHelper::lookupTransformChain(*segment, query);
            if (not result)
            {
                // TODO
                continue;
            }

            frames.emplace(robotName, result.transforms);
        }

        ARMARX_INFO << deactivateSpam(60)
                    << "Number of known robot pose chains: " << frames.size();

        return frames;
    }


    Segment::RobotPoseMap
    Segment::getRobotGlobalPosesLocking(const armem::Time& timestamp) const
    {
        std::scoped_lock lock(mutex());
        return getRobotGlobalPoses(timestamp);
    }


    Segment::RobotPoseMap
    Segment::getRobotGlobalPoses(const armem::Time& timestamp) const
    {
        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        RobotPoseMap robotGlobalPoses;

        const std::vector<std::string> knownRobots = simox::alg::get_keys(*segment);

        for (const std::string& robotName : knownRobots)
        {

            TransformQuery query{.header{.parentFrame = GlobalFrame,
                                         .frame       = "root", // TODO, FIXME
                                         .agent       = robotName,
                                         .timestamp   = timestamp}};

            const auto result = TransformHelper::lookupTransform(*segment, query);
            if (not result)
            {
                // TODO
                continue;
            }

            robotGlobalPoses.emplace(robotName, result.transform.transform);
        }

        ARMARX_INFO << deactivateSpam(60)
                    << "Number of known robot poses: " << robotGlobalPoses.size();

        return robotGlobalPoses;
    }


    bool Segment::commitTransform(const armarx::armem::robot_state::Transform& transform)
    {
        Commit commit;
        commit.add(makeUpdate(transform));

        const armem::CommitResult result = iceMemory.commit(commit);
        return result.allSuccess();
    }


    bool Segment::commitTransformLocking(const armarx::armem::robot_state::Transform& transform)
    {
        Commit commit;
        commit.add(makeUpdate(transform));

        const armem::CommitResult result = iceMemory.commitLocking(commit);
        return result.allSuccess();
    }


    EntityUpdate Segment::makeUpdate(const armarx::armem::robot_state::Transform& transform) const
    {
        const armem::Time& timestamp = transform.header.timestamp;
        const MemoryID providerID = segment->id().withProviderSegmentName(transform.header.agent);

        EntityUpdate update;
        update.entityID = providerID.withEntityName(transform.header.parentFrame + "," + transform.header.frame);
        update.timeArrived = update.timeCreated = update.timeSent = timestamp;

        arondto::Transform aronTransform;
        toAron(aronTransform, transform);
        update.instancesData = {aronTransform.toAron()};

        return update;
    }

} // namespace armarx::armem::server::robot_state::localization
