/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <vector>
#include <map>
#include <string>

// Simox
#include <SimoxUtility/meta/type_name.h>
#include <SimoxUtility/algorithm/string.h>
#include <SimoxUtility/algorithm/vector.hpp>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppBlock.h>
#include <ArmarXCore/libraries/cppgen/CppField.h>
#include <ArmarXCore/libraries/cppgen/CppCtor.h>
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>

#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/type/Navigator.h>

namespace armarx::aron::cppserializer
{
    class SerializerFactory;
    typedef std::shared_ptr<SerializerFactory> SerializerFactoryPtr;

    class Serializer;
    typedef std::shared_ptr<Serializer> SerializerPtr;

    class Serializer
    {
    public:
        using PointerType = SerializerPtr;

    public:
        // constructors
        Serializer() = default;
        Serializer(const std::string& cppName, const std::string& aronDataTypename, const std::string& aronTypeTypename);

        // public member methods
        std::string getCoreCppTypename() const;
        virtual std::string getFullCppTypename() const = 0;
        virtual std::string getFullTypenameGenerator() const = 0;

        std::string getAronDataTypename() const;
        std::string getAronDataPtrTypename() const;

        std::string getAronTypeTypename() const;
        std::string getAronTypePtrTypename() const;

        CppMethodPtr toSpecializedDataWriterMethod(const std::string&, const std::string&, const std::string&, const std::string& enforceType = "") const;
        CppMethodPtr toSpecializedDataReaderMethod(const std::string&, const std::string&, const std::string&, const std::string& enforceType = "") const;

        CppMethodPtr toSpecializedTypeWriterMethod(const std::string&, const std::string&, const std::string&, const std::string& enforceType = "") const;

        // virtual override definitions
        virtual std::vector<CppFieldPtr> getPublicVariableDeclarations(const std::string&) const = 0;

        CppCtorPtr toCtor(const std::string&) const;
        virtual std::vector<std::pair<std::string, std::string>> getCtorInitializers(const std::string&) const = 0;
        virtual CppBlockPtr getCtorBlock(const std::string&) const = 0;

        CppMethodPtr toResetSoftMethod() const;
        virtual CppBlockPtr getResetSoftBlock(const std::string&) const = 0;

        CppMethodPtr toWriteTypeMethod() const;
        virtual CppBlockPtr getWriteTypeBlock(const std::string&) const = 0;

        CppMethodPtr toResetHardMethod() const;
        virtual CppBlockPtr getResetHardBlock(const std::string&) const = 0;

        CppMethodPtr toWriteMethod() const;
        virtual CppBlockPtr getWriteBlock(const std::string&) const = 0;

        CppMethodPtr toReadMethod() const;
        virtual CppBlockPtr getReadBlock(const std::string&) const = 0;

        CppMethodPtr toEqualsMethod() const;
        virtual CppBlockPtr getEqualsBlock(const std::string&, const std::string&) const = 0;

        // static methods
        static std::string EscapeAccessor(const std::string&);
        static std::string UnescapeAccessor(const std::string&);

        static std::string ExtractCppTypename(const typenavigator::NavigatorPtr&);
        static std::vector<std::string> ExtractCppTypenames(const std::vector<typenavigator::NavigatorPtr>&);

        static SerializerPtr FromAronTypeNaviagtorPtr(const typenavigator::NavigatorPtr&);

        static std::string ResolveMaybeAccessor(const std::string&, const typenavigator::NavigatorPtr&);
        static CppBlockPtr ResolveMaybeWriteBlock(const std::string&, const CppBlockPtr&, const typenavigator::NavigatorPtr&);
        static CppBlockPtr ResolveMaybeReadBlock(const std::string&, const std::string&, const CppBlockPtr&, const typenavigator::NavigatorPtr&, bool enforce_return_type = false);
        static CppBlockPtr ResolveMaybeEqualsBlock(const std::string&, const std::string&, const CppBlockPtr&, const typenavigator::NavigatorPtr&);
        static CppBlockPtr ResolveMaybeResetSoftBlock(const std::string&, const CppBlockPtr&, const typenavigator::NavigatorPtr&);
        static CppBlockPtr ResolveMaybeResetHardBlock(const std::string&, const CppBlockPtr&, const typenavigator::NavigatorPtr&);

    protected:
        static const std::string READ_START_RETURN_TYPE_ACCESSOR;

    private:
        const static std::string ARON_DATA_NAME;
        const static std::string ARON_DATA_PTR_NAME;

        const static std::string ARON_TYPE_NAME;
        const static std::string ARON_TYPE_PTR_NAME;

        static const std::map<std::string, std::string> ESCAPE_ACCESSORS;

        static const SerializerFactoryPtr FACTORY;

        std::string cppTypename;
        std::string aronDataTypename;
        std::string aronTypeTypename;
    };
}
