/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL

// Simox
#include <SimoxUtility/algorithm/string.h>

// Header
#include "ReaderFactory.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/codegenerator/typeReader/xml/Data.h>

namespace armarx::aron::xmltypereader
{
    std::stack<codegeneratorhelper::GenerateObjectInfoPtr> ReaderFactory::ObjectGenerationStack = {};
    std::map<std::string, codegeneratorhelper::GenerateObjectInfoPtr> ReaderFactory::AllPublicObjects = {};
    std::map<std::string, codegeneratorhelper::GenerateIntEnumInfoPtr> ReaderFactory::AllPublicIntEnums = {};


    typenavigator::NavigatorPtr ReaderFactory::create(const RapidXmlReaderNode& node, const Path& path) const
    {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    {Data::GENERATE_##capsType##_MEMBER_TAG, ReaderFactoryPtr(new upperType##ReaderFactory())},
        static const std::map<std::string, ReaderFactoryPtr> Factories =
        {
            {Data::GENERATE_OBJECT_TAG, ReaderFactoryPtr(new ObjectReaderFactory())},
            {Data::GENERATE_INT_ENUM_TAG, ReaderFactoryPtr(new IntEnumReaderFactory())},
            HANDLE_ALL_ARON_TYPES
        };
#undef RUN_ARON_MACRO

        const std::string tag = simox::alg::to_lower(node.name());
        if (tag == "")
        {
            throw error::AronException("XMLReaderFactory", "create", "An node is empty. Cannot find facory for empty nodes.");
        }
        auto factory_iterator = Factories.find(tag);
        if (factory_iterator != Factories.end())
        {
            return factory_iterator->second->createSpecific(node, path);
        }

        // Search for known objects
        return ResolveTypename(tag);
    }

    typenavigator::NavigatorPtr ReaderFactory::createSpecific(const RapidXmlReaderNode&, const Path& p) const
    {
        throw error::AronException("ReaderFactory", "createSpecific", "Called disallowed method of an NavigatorFactory. Use child class instead!", p);
    }

    typenavigator::NavigatorPtr ReaderFactory::ResolveTypename(const std::string& name)
    {
        const auto public_intenum_it = AllPublicIntEnums.find(name);
        if (public_intenum_it != AllPublicIntEnums.end())
        {
            // copy the navigator
            return typenavigator::Navigator::FromAronType(public_intenum_it->second->correspondingType->toAronTypePtr());
        }

        const auto public_obj_it = AllPublicObjects.find(name);
        if (public_obj_it != AllPublicObjects.end())
        {
            // copy the navigator
            return typenavigator::Navigator::FromAronType(public_obj_it->second->correspondingType->toAronTypePtr());
        }

        throw error::StringNotValidException("XMLReaderFactory", "ResolveTypename", "Could not find a tag. You can only access already finished public objects.", name);
    }

    type::Maybe ReaderFactory::GetMaybe(const RapidXmlReaderNode& n)
    {
        if (Data::AttributeIsTrue(n, Data::OPTIONAL_NAME))
        {
            return type::Maybe::eOptional;
        }
        if (Data::AttributeIsTrue(n, Data::RAW_PTR_NAME))
        {
            return type::Maybe::eRawPointer;
        }
        if (Data::AttributeIsTrue(n, Data::SHARED_PTR_NAME))
        {
            return type::Maybe::eSharedPointer;
        }
        if (Data::AttributeIsTrue(n, Data::UNIQUE_PTR_NAME))
        {
            return type::Maybe::eUniquePointer;
        }
        return type::Maybe::eNone;
    }

    // Object type
    typenavigator::NavigatorPtr ObjectReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceAttribute(node, Data::NAME_ATTRIBUTE_NAME);
        const std::string extends = Data::GetAttributeWithDefault(node, Data::EXTENDS_ATTRIBUTE_NAME, "");
        const std::string name = Data::GetAttribute(node, Data::NAME_ATTRIBUTE_NAME);

        bool object_is_nested = ObjectGenerationStack.size() > 0;
        if (object_is_nested)
        {
            if (name.rfind("::") != std::string::npos)
            {
                throw error::StringNotValidException("AronObjectTypeXMLReaderFactory", "createSpecific", "Namespaces are not allowed in nested classes", name, path);
            }
        }

        // set the new object
        auto aronObjectType = std::make_shared<typenavigator::ObjectNavigator>(path);
        aronObjectType->setObjectName(name);

        auto newObject = std::make_shared<codegeneratorhelper::GenerateObjectInfo>();
        newObject->typeName = name;
        newObject->correspondingType = aronObjectType;
        newObject->doc_brief = Data::GetAttributeWithDefault(node, Data::DOC_BRIEF_NAME, "");
        newObject->doc_author = Data::GetAttributeWithDefault(node, Data::DOC_AUTHOR_NAME, "");

        if (extends != "")
        {
            auto parentObj = typenavigator::ObjectNavigator::DynamicCastAndCheck(ResolveTypename(simox::alg::to_lower(extends)));
            aronObjectType->setExtends(parentObj);
        }

        ObjectGenerationStack.push(newObject);
        for (const RapidXmlReaderNode& objectChild : node.nodes())
        {
            Data::EnforceTagName(objectChild, Data::OBJECT_CHILD_TAG);
            Data::EnforceChildSize(objectChild, 1);

            Data::EnforceAttribute(objectChild, Data::KEY_ATTRIBUTE_NAME);
            const std::string key = objectChild.attribute_value(Data::KEY_ATTRIBUTE_NAME);

            if (Data::HasAttribute(objectChild, Data::DOC_BRIEF_NAME))
            {
                newObject->doc_members.insert({key, Data::GetAttribute(objectChild, Data::DOC_BRIEF_NAME)});
            }

            std::vector<RapidXmlReaderNode> children = objectChild.nodes();

            auto maybe = GetMaybe(children[0]);
            typenavigator::NavigatorPtr childNavigator = create(children[0], Path(path, key));

            childNavigator->setMaybe(maybe);
            aronObjectType->addMemberType(key, childNavigator);
        }
        ObjectGenerationStack.pop();

        if (object_is_nested)
        {
            // We are in a local nested class since there is a parent class in stack
            codegeneratorhelper::GenerateObjectInfoPtr parent = ObjectGenerationStack.top();
            parent->nestedObjects.insert(std::make_pair(simox::alg::to_lower(newObject->typeName), newObject));
        }
        else
        {
            // We are a public top level class. Add to AllPublicObjects
            AllPublicObjects.insert(std::make_pair(simox::alg::to_lower(newObject->typeName), newObject));
            aronObjectType = typenavigator::ObjectNavigator::DynamicCastAndCheck(ResolveTypename(simox::alg::to_lower(newObject->typeName))); // make a copy
        }

        ARMARX_CHECK_NOT_NULL(aronObjectType);
        return aronObjectType;
    }

    // List type (List)
    typenavigator::NavigatorPtr ListReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_LIST_MEMBER_TAG);

        typenavigator::ListNavigatorPtr list(new typenavigator::ListNavigator(path));
        Data::EnforceChildSize(node, 1);
        std::vector<RapidXmlReaderNode> listTypeNodeChildren = node.nodes();
        const RapidXmlReaderNode typeNode = listTypeNodeChildren[0];
        auto maybe = GetMaybe(typeNode);
        typenavigator::NavigatorPtr type = create(typeNode, Path(path, "type"));
        type->setMaybe(maybe);

        list->setAcceptedType(type);
        return list;
    }

    // List type (Dict)
    typenavigator::NavigatorPtr DictReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_DICT_MEMBER_TAG);

        typenavigator::DictNavigatorPtr dict(new typenavigator::DictNavigator(path));
        Data::EnforceChildSize(node, 1);
        std::vector<RapidXmlReaderNode> dictTypeNodeChildren = node.nodes();
        const RapidXmlReaderNode typeNode = dictTypeNodeChildren[0];
        auto maybe = GetMaybe(typeNode);
        typenavigator::NavigatorPtr type = create(typeNode, Path(path, "type"));
        type->setMaybe(maybe);

        dict->setAcceptedType(type);
        return dict;
    }

    // List type (Tuple)
    typenavigator::NavigatorPtr TupleReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_TUPLE_MEMBER_TAG);
        std::vector<RapidXmlReaderNode> nodeChildren = node.nodes();

        if (nodeChildren.size() < 1)
        {
            throw error::AronException("AronTupleTypeXMLReaderFactory", "createSpecific", "A tuple must have at least one type. Found child nodes size: " + std::to_string(nodeChildren.size()));
        }

        typenavigator::TupleNavigatorPtr tuple(new typenavigator::TupleNavigator(path));
        unsigned int i = 0;
        for (const RapidXmlReaderNode& tupleTypeDeclarationNode : nodeChildren)
        {
            Data::EnforceChildSize(tupleTypeDeclarationNode, 1);

            std::vector<RapidXmlReaderNode> typeNodeChildren = tupleTypeDeclarationNode.nodes();
            const RapidXmlReaderNode typeNode = typeNodeChildren[0];
            auto maybe = GetMaybe(typeNode);

            typenavigator::NavigatorPtr type = create(typeNode, Path(path, "<" + std::to_string(i++) + ">"));
            type->setMaybe(maybe);

            tuple->addAcceptedType(type);
        }
        return tuple;
    }

    // List type (Pair)
    typenavigator::NavigatorPtr PairReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_PAIR_MEMBER_TAG);
        std::vector<RapidXmlReaderNode> nodeChildren = node.nodes();

        if (nodeChildren.size() != 2)
        {
            throw error::AronException("AronPairTypeXMLReaderFactory", "createSpecific", "A pair must have two types. Found child nodes size: " + std::to_string(nodeChildren.size()));
        }

        typenavigator::PairNavigatorPtr pair(new typenavigator::PairNavigator(path));

        Data::EnforceChildSize(nodeChildren[0], 1);
        std::vector<RapidXmlReaderNode> type1NodeChildren = nodeChildren[0].nodes();
        const RapidXmlReaderNode type1Node = type1NodeChildren[0];

        auto maybe1 = GetMaybe(type1Node);

        typenavigator::NavigatorPtr type1 = create(type1Node, Path(path, std::to_string(0)));
        type1->setMaybe(maybe1);
        pair->setFirstAcceptedType(type1);

        Data::EnforceChildSize(nodeChildren[1], 1);
        std::vector<RapidXmlReaderNode> type2NodeChildren = nodeChildren[1].nodes();
        const RapidXmlReaderNode type2Node = type2NodeChildren[0];

        auto maybe2 = GetMaybe(type2Node);

        typenavigator::NavigatorPtr type2 = create(type2Node, Path(path, std::to_string(1)));
        type2->setMaybe(maybe2);
        pair->setSecondAcceptedType(type2);

        return pair;
    }

    // Complex type (NDArray)
    typenavigator::NavigatorPtr NDArrayReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        return nullptr;
    }

    // Complex type (IVTCByteImage)
    typenavigator::NavigatorPtr IVTCByteImageReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_IVT_CBYTE_IMAGE_MEMBER_TAG);
        typenavigator::IVTCByteImageNavigatorPtr complex(new typenavigator::IVTCByteImageNavigator(path));
        return complex;
    }

    // Complex type (EigenMatrix)
    typenavigator::NavigatorPtr EigenMatrixReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_EIGEN_MATRIX_MEMBER_TAG);
        Data::EnforceAttribute(node, Data::TYPE_ATTRIBUTE_NAME);

        const int rows = std::stoi(Data::GetAttributeWithDefault(node, Data::ROWS_ATTRIBUTE_NAME, "4"));
        const int cols = std::stoi(Data::GetAttributeWithDefault(node, Data::COLS_ATTRIBUTE_NAME, "4"));
        std::string type = node.attribute_value(Data::TYPE_ATTRIBUTE_NAME);

        typenavigator::EigenMatrixNavigatorPtr complex(new typenavigator::EigenMatrixNavigator(path));
        complex->setRows(rows);
        complex->setCols(cols);
        complex->setTypename(type);
        return complex;
    }

    // Complex type (EigenQuaternion)
    typenavigator::NavigatorPtr EigenQuaternionReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_EIGEN_QUATERNION_MEMBER_TAG);
        Data::EnforceAttribute(node, Data::TYPE_ATTRIBUTE_NAME);

        std::string type = node.attribute_value(Data::TYPE_ATTRIBUTE_NAME);

        typenavigator::EigenQuaternionNavigatorPtr complex(new typenavigator::EigenQuaternionNavigator(path));
        complex->setTypename(type);
        return complex;
    }

    // Complex type (OpenCVMat)
    typenavigator::NavigatorPtr OpenCVMatReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_OPENCV_MAT_MEMBER_TAG);

        typenavigator::OpenCVMatNavigatorPtr complex(new typenavigator::OpenCVMatNavigator(path));
        return complex;
    }

    // Complex type (PCLPointCloud)
    typenavigator::NavigatorPtr PCLPointCloudReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_PCL_POINTCLOUD_MEMBER_TAG);
        Data::EnforceAttribute(node, Data::TYPE_ATTRIBUTE_NAME);
        std::string type = node.attribute_value(Data::TYPE_ATTRIBUTE_NAME);;

        typenavigator::PCLPointCloudNavigatorPtr complex(new typenavigator::PCLPointCloudNavigator(path));
        complex->setTypename(type);
        return complex;
    }

    // Complex type (Position)
    typenavigator::NavigatorPtr PositionReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_POSITION_MEMBER_TAG);

        typenavigator::PositionNavigatorPtr complex(new typenavigator::PositionNavigator(path));
        return complex;
    }

    // Complex type (Orientation)
    typenavigator::NavigatorPtr OrientationReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_ORIENTATION_MEMBER_TAG);

        typenavigator::OrientationNavigatorPtr complex(new typenavigator::OrientationNavigator(path));
        return complex;
    }

    // Complex type (Pose)
    typenavigator::NavigatorPtr PoseReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceTagName(node, Data::GENERATE_POSE_MEMBER_TAG);

        typenavigator::PoseNavigatorPtr complex(new typenavigator::PoseNavigator(path));
        return complex;
    }



    // Enum (IntEnum)
    typenavigator::NavigatorPtr IntEnumReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const
    {
        Data::EnforceAttribute(node, Data::NAME_ATTRIBUTE_NAME);
        const std::string name = Data::GetAttribute(node, Data::NAME_ATTRIBUTE_NAME);

        bool enum_is_nested = ObjectGenerationStack.size() > 0;
        if (enum_is_nested)
        {
            if (name.rfind("::") != std::string::npos)
            {
                throw error::StringNotValidException("IntEnumReaderFactory", "createSpecific", "Namespaces are not allowed in nested enums", name, path);
            }
        }

        // set the new object
        auto aronIntEnumType = std::make_shared<typenavigator::IntEnumNavigator>(path);
        aronIntEnumType->setEnumName(name);

        auto newEnumInfo = std::make_shared<codegeneratorhelper::GenerateIntEnumInfo>();
        newEnumInfo->typeName = name;
        newEnumInfo->correspondingType = aronIntEnumType;

        for (const RapidXmlReaderNode& valueChild : node.nodes())
        {
            Data::EnforceTagName(valueChild, Data::ENUM_VALUE_TAG);
            Data::EnforceChildSize(valueChild, 0);

            Data::EnforceAttribute(valueChild, Data::KEY_ATTRIBUTE_NAME);
            const std::string key = valueChild.attribute_value(Data::KEY_ATTRIBUTE_NAME);

            if (Data::HasAttribute(valueChild, Data::DOC_BRIEF_NAME))
            {
                newEnumInfo->doc_values.insert({key, Data::GetAttribute(valueChild, Data::DOC_BRIEF_NAME)});
            }

            Data::EnforceAttribute(valueChild, Data::VALUE_ATTRIBUTE_NAME);
            const std::string value = valueChild.attribute_value(Data::VALUE_ATTRIBUTE_NAME);

            aronIntEnumType->addAcceptedValue(key, std::stoi(value));
        }

        if (enum_is_nested)
        {
            // We are in a local nested class since there is a parent class in stack
            codegeneratorhelper::GenerateObjectInfoPtr parent = ObjectGenerationStack.top();
            parent->nestedIntEnums.insert(std::make_pair(simox::alg::to_lower(newEnumInfo->typeName), newEnumInfo));
        }
        else
        {
            // We are a public top level class. Add to AllPublicObjects
            AllPublicIntEnums.insert(std::make_pair(simox::alg::to_lower(newEnumInfo->typeName), newEnumInfo));
        }

        return aronIntEnumType;
    }

    // Primitve Types
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    typenavigator::NavigatorPtr upperType##ReaderFactory::createSpecific(const RapidXmlReaderNode& node, const Path& path) const \
    { \
        Data::EnforceTagName(node, Data::GENERATE_##capsType##_MEMBER_TAG); \
        return typenavigator::NavigatorPtr(new typenavigator::upperType##Navigator(path)); \
    }

    HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
}
