/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller (fabian dot peller at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NlohmannJSONWriter.h"

namespace armarx::aron::typeIO::writer
{

    void NlohmannJSONWriter::writeStartObject(const WriteStartObjectInput& o)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_OBJECT_NAME_SLUG] = o.name;
        data[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eObject, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndObject()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartList(const WriteStartListInput& o)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eList, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartDict(const WriteStartDictInput& o)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eDict, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartTuple(const WriteStartTupleInput& o)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::eTuple, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndTuple()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartPair(const WriteStartPairInput& o)
    {
        nlohmann::json data;
        data[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(type::Descriptor::ePair, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndPair()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NlohmannJSONWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeEigenMatrix(const WriteEigenMatrixInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "EigenMatrix";
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = std::vector<unsigned int>({o.rows, o.cols});
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = o.type;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeEigenQuaternion(const WriteEigenQuaternionInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "EigenQuaternion";
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = o.type;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeIVTCByteImage(const WriteIVTCByteImageInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "IVTCByteImage";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeOpenCVMat(const WriteOpenCVMatInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "OpenCVMat";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePCLPointCloud(const WritePCLPointCloudInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "PCLPointCloud";
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = o.type;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePosition(const WritePositionInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Position";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeOrientation(const WriteOrientationInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Orientation";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePose(const WritePoseInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        j[io::Data::READER_WRITER_NDARRAY_NAME_SLUG] = "Pose";
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeInt(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_INT_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeLong(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_LONG_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeFloat(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_FLOAT_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeDouble(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_DOUBLE_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeString(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_STRING_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeBool(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_BOOL_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeTime(const WritePrimitiveInput& o)
    {
        auto token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_TYPE_SLUG] = io::Data::READER_WRITER_TIME_TYPENAME_SLUG;
        j[io::Data::READER_WRITER_MAYBE_SLUG] = o.maybe;
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    void NlohmannJSONWriter::writeExtends()
    {
        if (stack.size() < 2)
        {
            throw error::SizeNotValidException("NavigatorWriter", "writeExtends", "The stacksize must at least be 2 (one for base, one for derived)", stack.size(), 2);
        }

        auto base = stack.top();
        stack.pop();
        auto derived = stack.top();

        if (!base->getElement().is_object())
        {
            throw error::AronException("NlohmannJSONWriter", "writeExtends", "Could not inherit from a non-object.");
        }

        if (!derived->getElement().is_object())
        {
            throw error::AronException("NlohmannJSONWriter", "writeExtends", "Could not inherit a non-object.");
        }

        derived->getElement()[io::Data::READER_WRITER_OBJECT_EXTENDS_SLUG] = base->getElement();
    }
}
